/**
 * 
 */
package reseptit;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;


/**
 * <pre>
 *|------------------------------------------------------------------------|
 *| Luokan nimi:  Ainesosat                            | Avustajat:        |
 *|-------------------------------------------------------------------------
 *| Vastuualueet:                                      |                   |
 *|                                                    | - Ainesosa        |
 *| - Pitää listaa ainesosista, sekä osaa lisätä ja    |                   |
 *|  poistaa ainesosia                                 |                   |
 *| - Osaa etisä ainesosia                             |                   |
 *| - Lukee ja kirjoittaa ohjeista tiedostoon          |                   |
 *|-------------------------------------------------------------------------
 * </pre>
 * @author Omistaja
 * @version 17.3.2025
 *
 */
public class Ainesosat implements Iterable<Ainesosa>{
    
    private String tiedPerusNimi = "Ainesosat";
    private boolean muutettu = false;
    
    private final Collection<Ainesosa> alkiot  = new ArrayList<Ainesosa>();
    
    
    /**
     * Joku konstruktori
     */
    public Ainesosat() {
        
    }
    
    
    /**
     * Lisää uuden ainesosan tietorakenteeseen.  Ottaa ainesosan omistukseensa.
     * @param ain lisättävä ainesosa.  Huom tietorakenne muuttuu omistajaksi
     */
    public void lisaa(Ainesosa ain) {
        alkiot.add(ain);
        muutettu = true;
    }
    
    
    /**
     * Lukee ainesosat tiedostosta.
     * @param tied tiedoston nimen alkuosa
     * @throws SailoException jos lukeminen epäonnistuu
     * 
     * @example
     * <pre name="test">
     * #THROWS SailoException 
     * #import java.io.File;
     *  Ainesosat aineet = new Ainesosat();
     *  Ainesosa maito21 = new Ainesosa(); maito21.vastaaMaito(2);
     *  Ainesosa maito11 = new Ainesosa(); maito11.vastaaMaito(1);
     *  Ainesosa maito22 = new Ainesosa(); maito22.vastaaMaito(2); 
     *  Ainesosa maito12 = new Ainesosa(); maito12.vastaaMaito(1); 
     *  Ainesosa maito23 = new Ainesosa(); maito23.vastaaMaito(2); 
     *  String tiedNimi = "testiainesosat";
     *  File ftied = new File(tiedNimi+".dat");
     *  ftied.delete();
     *  aineet.lueTiedostosta(tiedNimi); #THROWS SailoException
     *  aineet.lisaa(maito21);
     *  aineet.lisaa(maito11);
     *  aineet.lisaa(maito22);
     *  aineet.lisaa(maito12);
     *  aineet.lisaa(maito23);
     *  aineet.talleta();
     *  aineet = new Ainesosat();
     *  aineet.lueTiedostosta(tiedNimi);
     *  Iterator<Ainesosa> i = aineet.iterator();
     *  i.next().toString() === maito21.toString();
     *  i.next().toString() === maito11.toString();
     *  i.next().toString() === maito22.toString();
     *  i.next().toString() === maito12.toString();
     *  i.next().toString() === maito23.toString();
     *  i.hasNext() === false;
     *  aineet.lisaa(maito23);
     *  aineet.talleta();
     *  ftied.delete() === true;
     *  File fbak = new File(tiedNimi+".bak");
     *  fbak.delete() === true;
     * </pre>
     * @throws SailoException jos lukeminen epäonnistuu
     */
    public void lueTiedostosta(String tied) throws SailoException {
        setTiedPerusNimi(tied);
        try ( BufferedReader fi = new BufferedReader(new FileReader(getTiedNimi())) ) {

            String rivi;
            while ( (rivi = fi.readLine()) != null ) {
                rivi = rivi.trim();
                if ( "".equals(rivi) || rivi.charAt(0) == ';' ) continue;
                Ainesosa ain = new Ainesosa();
                ain.parse(rivi); // voisi olla virhekäsittely
                lisaa(ain);
            }
            muutettu = false;

        } catch ( FileNotFoundException e ) {
            throw new SailoException("Tiedosto " + getTiedNimi() + " ei aukea");
        } catch ( IOException e ) {
            throw new SailoException("Ongelmia tiedoston kanssa: " + e.getMessage());
        }

    }
    
    
    /**
     * Luetaan aikaisemmin annetun nimisestä tiedostosta
     * @throws SailoException jos tulee poikkeus
     */
    public void lueTiedostosta() throws SailoException {
        lueTiedostosta(getTiedPerusNimi());
    }


    /**
     * Tallentaa ainesosat tiedostoon.
     * @throws SailoException jos talletus epäonnistuu
     */
    public void talleta() throws SailoException {
        if ( !muutettu ) return;

        File fbak = new File(getBakNimi());
        File ftied = new File(getTiedNimi());
        fbak.delete(); //  if ... System.err.println("Ei voi tuhota");
        ftied.renameTo(fbak); //  if ... System.err.println("Ei voi nimetä");

        try ( PrintWriter fo = new PrintWriter(new FileWriter(ftied.getCanonicalPath())) ) {
            for (Ainesosa ain : this) {
                fo.println(ain.toString());
            }
        } catch ( FileNotFoundException ex ) {
            throw new SailoException("Tiedosto " + ftied.getName() + " ei aukea");
        } catch ( IOException ex ) {
            throw new SailoException("Tiedoston " + ftied.getName() + " kirjoittamisessa ongelmia");
        }

        muutettu = false;

    }
    
    
    /**
     * Asettaa tiedoston perusnimen ilman tarkenninta
     * @param tied tallennustiedoston perusnimi
     */
    public void setTiedPerusNimi(String tied) {
        tiedPerusNimi = tied;
    }


    /**
     * Palauttaa tiedoston nimen, jota käytetään tallennukseen
     * @return tallennustiedoston nimi
     */
    public String getTiedPerusNimi() {
        return tiedPerusNimi;
    }


    /**
     * Palauttaa tiedoston nimen, jota käytetään tallennukseen
     * @return tallennustiedoston nimi
     */
    public String getTiedNimi() {
        return tiedPerusNimi + ".dat";
    }


    /**
     * Palauttaa varakopiotiedoston nimen
     * @return varakopiotiedoston nimi
     */
    public String getBakNimi() {
        return tiedPerusNimi + ".bak";
    }


    /**
     * Palauttaa Ainesosien lukumäärän
     * @return ainesosien lukumäärä
     */
    public int getALkm() {
        return alkiot.size();
    }

    
    /**
     * Iteraattori kaikkien ainesosien läpikäymiseen
     * @return ainesosaiteraattori
     * 
     * @example
     * <pre name="test">
     * #PACKAGEIMPORT
     * #import java.util.*;
     * 
     *  Ainesosat aineet = new Ainesosat();
     *  Ainesosa maito21 = new Ainesosa(2); aineet.lisaa(maito21);
     *  Ainesosa maito11 = new Ainesosa(1); aineet.lisaa(maito11);
     *  Ainesosa maito22 = new Ainesosa(2); aineet.lisaa(maito22);
     *  Ainesosa maito12 = new Ainesosa(1); aineet.lisaa(maito12);
     *  Ainesosa maito23 = new Ainesosa(2); aineet.lisaa(maito23);
     * 
     *  Iterator<Ainesosa> i2=aineet.iterator();
     *  i2.next() === maito21;
     *  i2.next() === maito11;
     *  i2.next() === maito22;
     *  i2.next() === maito12;
     *  i2.next() === maito23;
     *  i2.next() === maito12;  #THROWS NoSuchElementException  
     *  
     *  int n = 0;
     *  int anrot[] = {2,1,2,1,2};
     *  
     *  for ( Ainesosa ain:aineet ) { 
     *    ain.getOhjeNro() === anrot[n]; n++;  
     *  }
     *  
     *  n === 5;
     *  
     * </pre>
     */
    @Override
    public Iterator<Ainesosa> iterator() {
        return alkiot.iterator();
    }
    
    
    /**
     * Haetaan kaikki ohjeen ainesosat
     * @param tunnusnro ohjeen tunnusnumero jolle ainesosia haetaan
     * @return tietorakenne jossa viiteet löydetteyihin ainesosiin
     * @example
     * <pre name="test">
     * #import java.util.*;
     * 
     *  Ainesosat aineet = new Ainesosat();
     *  Ainesosa maito21 = new Ainesosa(2); aineet.lisaa(maito21);
     *  Ainesosa maito11 = new Ainesosa(1); aineet.lisaa(maito11);
     *  Ainesosa maito22 = new Ainesosa(2); aineet.lisaa(maito22);
     *  Ainesosa maito12 = new Ainesosa(1); aineet.lisaa(maito12);
     *  Ainesosa maito23 = new Ainesosa(2); aineet.lisaa(maito23);
     *  Ainesosa maito51 = new Ainesosa(5); aineet.lisaa(maito51);
     *  
     *  List<Ainesosa> loytyneet;
     *  loytyneet = aineet.annaAinesosat(3);
     *  loytyneet.size() === 0; 
     *  loytyneet = aineet.annaAinesosat(1);
     *  loytyneet.size() === 2; 
     *  loytyneet.get(0) == maito11 === true;
     *  loytyneet.get(1) == maito12 === true;
     *  loytyneet = aineet.annaAinesosat(5);
     *  loytyneet.size() === 1; 
     *  loytyneet.get(0) == maito51 === true;
     * </pre> 
     */
    public List<Ainesosa> annaAinesosat(int tunnusnro) {
        List<Ainesosa> loydetyt = new ArrayList<Ainesosa>();
        for (Ainesosa ain : alkiot)
            if (ain.getOhjeNro() == tunnusnro) loydetyt.add(ain);
        return loydetyt;
    }

    
    /**
     * Poistaa kaikki tietyn tietyn ohjeen ainesosat
     * @param tunnusNro viite siihen, mihin liittyvät tietueet poistetaan
     * @return montako poistettiin 
     * @example
     * <pre name="test">
     *  Ainesosat aineet = new Ainesosat();
     *  Ainesosa maito21 = new Ainesosa(); maito21.vastaaMaito(2);
     *  Ainesosa maito11 = new Ainesosa(); maito11.vastaaMaito(1);
     *  Ainesosa maito22 = new Ainesosa(); maito22.vastaaMaito(2); 
     *  Ainesosa maito12 = new Ainesosa(); maito12.vastaaMaito(1); 
     *  Ainesosa maito23 = new Ainesosa(); maito23.vastaaMaito(2); 
     *  aineet.lisaa(maito21);
     *  aineet.lisaa(maito11);
     *  aineet.lisaa(maito22);
     *  aineet.lisaa(maito12);
     *  aineet.lisaa(maito23);
     *  aineet.poistaOhjeenAinesosat(2) === 3;  aineet.getLkm() === 2;
     *  aineet.poistaOhjeenAinesosat(3) === 0;  aineet.getLkm() === 2;
     *  List<Ainesosa> h = aineet.annaAinesosat(2);
     *  h.size() === 0; 
     *  h = aineet.annaAinesosat(1);
     *  h.get(0) === maito11;
     *  h.get(1) === maito12;
     * </pre>
     */
    public int poistaOhjeenAinesosat(int tunnusNro) {
        int n = 0;
        for (Iterator<Ainesosa> it = alkiot.iterator(); it.hasNext();) {
            Ainesosa ain = it.next();
            if ( ain.getOhjeNro() == tunnusNro ) {
                it.remove();
                n++;
            }
        }
        if (n > 0) muutettu = true;
        return n;
    }
    
    
    /**
     * Poistaa valitun ainesosan
     * @param ainesosa poistettava ainesosa
     * @return tosi jos löytyi poistettava tietue 
     * @example
     * <pre name="test">
     * #THROWS SailoException 
     * #import java.io.File;
     *  Ainesosat aineet = new Ainesosat();
     *  Ainesosa maito21 = new Ainesosa(); maito21.vastaaMaito(2);
     *  Ainesosa maito11 = new Ainesosa(); maito11.vastaaMaito(1);
     *  Ainesosa maito22 = new Ainesosa(); maito22.vastaaMaito(2); 
     *  Ainesosa maito12 = new Ainesosa(); maito12.vastaaMaito(1); 
     *  Ainesosa maito23 = new Ainesosa(); maito23.vastaaMaito(2); 
     *  aineet.lisaa(maito21);
     *  aineet.lisaa(maito11);
     *  aineet.lisaa(maito22);
     *  aineet.lisaa(maito12);
     *  aineet.poista(maito23) === false ; aineet.getLkm() === 4;
     *  aineet.poista(maito11) === true;   aineet.getLkm() === 3;
     *  List<Ainesosa> h = aineet.annaAinesosat(1);
     *  h.size() === 1; 
     *  h.get(0) === maito12;
     * </pre>
     */
    public boolean poista(Ainesosa ainesosa) {
        boolean ret = alkiot.remove(ainesosa);
        if (ret) muutettu = true;
        return ret;
    }


        
    /**
     * Malliksi ainesosalista
     * @return malliainesosalista 
     */
    public List<Ainesosa> MalliAinesosat1() {
        List<Ainesosa> aineet1 = new ArrayList<Ainesosa>();
         Ainesosa aine10 = new Ainesosa();
         aine10.vastaaVehnajauho(1);
         aineet1.add(aine10);
         Ainesosa aine11 = new Ainesosa();
         aine11.vastaaMaito(1);
         aineet1.add(aine11);
         return aineet1;
    }
    
    
    /**
     * Malliksi toinen ainesosalista
     * @return toinen malliainesosalista
     */
    public List<Ainesosa> MalliAinesosat2() {
        List <Ainesosa> aineet2 = new ArrayList<Ainesosa>();
        Ainesosa aine20 = new Ainesosa();
        aine20.vastaaMaito(2);
        aineet2.add(aine20);
        Ainesosa aine21 = new Ainesosa();
        aine21.vastaaSuola(2);
        aineet2.add(aine21);
        return aineet2;
    }

    
    /**
     * @param args ei käytössä
     */
    public static void main(String[] args) {
        Ainesosat aineet = new Ainesosat();
        Ainesosa maito1 = new Ainesosa();
        maito1.vastaaMaito(2);
        Ainesosa maito2 = new Ainesosa();
        maito2.vastaaMaito(1);
        Ainesosa maito3 = new Ainesosa();
        maito3.vastaaMaito(2);
        Ainesosa maito4 = new Ainesosa();
        maito4.vastaaMaito(2);

        aineet.lisaa(maito1);
        aineet.lisaa(maito2);
        aineet.lisaa(maito3);
        aineet.lisaa(maito2);
        aineet.lisaa(maito4);

        System.out.println("============= Ainesosat testi =================");

        List<Ainesosa> Ainesosat2 = aineet.annaAinesosat(2);

        for (Ainesosa ain : Ainesosat2) {
            System.out.print(ain.getId() + " ");
            ain.tulosta(System.out);
        }
    
    }

}
